"""Django App Search Client utilities."""

from django.apps import apps
from elastic_app_search import Client


def get_api_v1_client():
    """Return the app search client."""
    config = apps.get_app_config('django_elastic_appsearch')

    base_endpoint = config.api_v1_base_endpoint
    api_key = config.api_key
    use_https = config.use_https

    return Client(
        api_key=api_key,
        base_endpoint=base_endpoint,
        use_https=use_https
    )



"""A Queryset slicer for Django."""

def slice_queryset(queryset, chunk_size):
    """ Slice a queryset into chunks. """

    start_pk = 0
    queryset = queryset.order_by('pk')

    while True:
        # No entry left
        if not queryset.filter(pk__gt=start_pk).exists():
            break

        try:
            # Fetch chunk_size entries if possible
            end_pk = queryset.filter(pk__gt=start_pk).values_list(
                'pk', flat=True)[chunk_size - 1]

            # Fetch rest entries if less than chunk_size left
        except IndexError:
            end_pk = queryset.values_list('pk', flat=True).last()

        yield queryset.filter(pk__gt=start_pk).filter(pk__lte=end_pk)

        start_pk = end_pk



"""Serialisers for Django models -> Elastic App Search objects."""

# pylint:disable=unused-import
from serpy import Field, MethodField, Serializer, StrField


class AppSearchSerialiser(Serializer):
    """A base serialiser to serialise Django models to app search objects."""

    id = MethodField()
    object_type = MethodField()

    def get_id(self, instance):
        """Set model pk as the document id."""
        return instance.get_appsearch_document_id()

    def get_object_type(self, instance):
        """Set model name as the object type."""
        return type(instance).__name__

# pylint:enable=unused-import



"""Django connector for Elastic App Search."""

# pylint:disable=invalid-name
# Django requires `default_app_config`
default_app_config = 'django_elastic_appsearch.apps.DjangoAppSearchConfig'
__version__ = '0.5.5'



"""ORM features for Elastic App Search."""

from django.apps import apps
from django.db import models

from django_elastic_appsearch.clients import get_api_v1_client
from django_elastic_appsearch.slicer import slice_queryset


class AppSearchQuerySet(models.QuerySet):
    """A queryset that supports Elastic App Search functions."""

    def _get_sliced_queryset(self):
        """Return the sliced queryset."""
        chunk_size = apps.get_app_config('django_elastic_appsearch').chunk_size
        return slice_queryset(self, chunk_size)

    def delete_from_appsearch(self):
        """Delete from appsearch."""
        if self and apps.get_app_config('django_elastic_appsearch').enabled:
            engine_name = self.first().get_appsearch_engine_name()
            client = self.first().get_appsearch_client()
            slices = self._get_sliced_queryset()
            for queryset in slices:
                client.destroy_documents(
                    engine_name,
                    [item.get_appsearch_document_id() for item in queryset]
                )

    def index_to_appsearch(self):
        """Index the queryset."""
        if self and apps.get_app_config('django_elastic_appsearch').enabled:
            engine_name = self.first().get_appsearch_engine_name()
            client = self.first().get_appsearch_client()
            slices = self._get_sliced_queryset()
            for queryset in slices:
                client.index_documents(
                    engine_name,
                    [item.serialise_for_appsearch() for item in queryset]
                )


class AppSearchModel(models.Model):
    """A model that integrates with Elastic App Search."""

    objects = AppSearchQuerySet.as_manager()

    class Meta:
        """Meta options for the app search model."""

        abstract = True

    @classmethod
    def get_appsearch_serialiser_class(cls):
        """Get the app search serialiser class."""
        return cls.AppsearchMeta.appsearch_serialiser_class

    @classmethod
    def get_appsearch_engine_name(cls):
        """Get the app search engine name that maps to this model."""
        return cls.AppsearchMeta.appsearch_engine_name or cls.__name__

    @classmethod
    def set_appsearch_serialiser_class(cls, serialiser_class):
        """Set the app search serialiser class."""
        cls.AppsearchMeta.appsearch_serialiser_class = serialiser_class

    @classmethod
    def set_appsearch_engine_name(cls, engine_name):
        """Set the app search engine name that maps to this model."""
        cls.AppsearchMeta.appsearch_engine_name = engine_name

    @classmethod
    def get_appsearch_client(cls):
        """Get the App Search client."""
        return get_api_v1_client()

    def serialise_for_appsearch(self):
        """Serialise the instance for appsearch."""
        _serialiser = self.get_appsearch_serialiser_class()
        return _serialiser(self).data

    def get_appsearch_document_id(self):
        """Get the unique document ID."""
        return '{}_{}'.format(type(self).__name__, self.pk)

    def index_to_appsearch(self):
        """Index the object to appsearch."""
        if apps.get_app_config('django_elastic_appsearch').enabled:
            self.get_appsearch_client().index_documents(
                self.get_appsearch_engine_name(),
                [self.serialise_for_appsearch()]
            )

    def delete_from_appsearch(self):
        """Delete the object from appsearch."""
        if apps.get_app_config('django_elastic_appsearch').enabled:
            self.get_appsearch_client().destroy_documents(
                self.get_appsearch_engine_name(),
                [self.get_appsearch_document_id()]
            )
